import weakref


class Event(object):
    def __init__(self, name, **kw):
        self.name = name
        self.__dict__.update(kw)
        self.method_name = 'EVT_%s' % self.name
    
    def __repr__(self):
        return "<Event %s %s>" % (self.name, self.args)
 
class Leaf(object):
    def dispatch(self, event):
        method = getattr(self, event.method_name, None)
        if method is not None: method(event)


class Node(object):
    def __init__(self):
        self.parent = None
        self.leaves = []
        self.branches = []
        
    def add(self, other):
        if isinstance(other, Node):
            other.parent = weakref.ref(self)
            self.branches.append(other)
        else:
            if hasattr(other, 'dispatch'):
                other.parent = weakref.ref(self)
                self.leaves.append(other)
            else:
                raise ValueError('Other object does not provide the dispatch protocol.')
    
    def dispatch(self, event):
        method_name = event.method_name
        for leaf in self.leaves:
            leaf.dispatch(event)
        method = getattr(self, method_name, None)
        if method is not None: 
            return method(event)
        else:
            self.dispatch_to_parent(event)
    
    def dispatch_to_parent(self, event):
        if self.parent is not None:
            parent = self.parent()
            dispatch = getattr(parent, 'dispatch', None)
            if dispatch is not None: 
                return dispatch(event)
        else:
            self.parent = None
            

class Root(Node):
    def __init__(self):
        Node.__init__(self)
        self.focused_node = self
    
    def focus(self, node):
        if node is self.focused_node: return
        self.focused_node.dispatch(Event('BLUR'))
        self.focused_node = node
        node.dispatch(Event('FOCUS'))
        
    def blur(self, node):
        self.focused_node = self
        node.dispatch(Event('BLUR'))
    
    def post(self, event_name, **kw):
        event = Event(event_name, **kw)
        return self.focused_node.dispatch(event)
        


if __name__ == "__main__":
    class Button(Node):
        def EVT_KeyDown(self, event):
            return self
            
    r = Root()
    b = Button()
    r.add(b)
    r.focus(b)
    
    print r.post('KeyDown')
    print r.post('Quit')
        





